#!/bin/sh
# ---------------------------------------------------------------------------
# JOnAS: Java(TM) Open Application Server
# Copyright (C) 2013 Bull S.A.S.
# Contact: jonas-team@ow2.org
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA
#
# Initial developer(s): Philippe Durieux
# Contributor(s): Adriana Danes :
# - Change /config en /conf
# - Use JONAS_BASE
# Florent Benoit & Ludovic Bert
# eyindanga
# - wrapper integration
# Jeremy Cazaux
# ---------------------------------------------------------------------------
# $Id$
# ---------------------------------------------------------------------------

cygwin=false;
case "`uname`" in
  CYGWIN*) cygwin=true ;;
esac

# For Cygwin, ensure paths are in UNIX format before anything is touched
if $cygwin ; then
  [ -n "JONAS_ROOT" ] &&
    JONAS_ROOT=`cygpath --unix "$JONAS_ROOT"`
  [ -n "JONAS_BASE" ] &&
    JONAS_BASE=`cygpath --unix "$JONAS_BASE"`
  [ -n "$JAVA_HOME" ] &&
    JAVA_HOME=`cygpath -m -s "$JAVA_HOME"`
fi

# -----------------------------------------------------------------------
# Compute JONAS_ROOT in case it has not been set.
# -----------------------------------------------------------------------
if [ -z "$JONAS_ROOT" ]; then
  savewd=`pwd`
  cd `dirname "$0"`
  zdir=`pwd`
  cd "$savewd"
  zfqn=$zdir/`basename "$0"`
  JONAS_ROOT=`dirname "$zdir"`
  export JONAS_ROOT
fi


# Check if $JONAS_ROOT is set.
if [ ! -f "$JONAS_ROOT"/bin/setenv ]; then
    echo "Please set the JONAS_ROOT variable."
    exit 1
fi

# Set CLASSPATH and JONAS_BASE
. "$JONAS_ROOT"/bin/setenv

# Display banner (if any)
if [ -f "$JONAS_BASE"/conf/banner.txt ]; then
    cat "$JONAS_BASE"/conf/banner.txt
fi

# For Cygwin, switch paths to Windows format before running java
if $cygwin; then
  [ -n "$JONAS_ROOT" ] &&
    JONAS_ROOT=`cygpath --path --windows "$JONAS_ROOT"`
  [ -n "$JONAS_BASE" ] &&
    JONAS_BASE=`cygpath --path --windows "$JONAS_BASE"`
fi

#-------------------------------------------------------
# Compute JONAS_WRAPPER_ROOT in case it has not been set.
#-------------------------------------------------------
if [ -z  "$JONAS_WRAPPER_ROOT" ]; then
  JONAS_WRAPPER_ROOT=${JONAS_ROOT}/wrapper
fi

if [ ! -d "$JONAS_WRAPPER_ROOT" ]; then
  echo "Please set the JONAS_WRAPPER_ROOT variable or put the wrapper root directory in $JONAS_ROOT/."
  exit 1
fi

#-----------------------------------------------------
# Compute JONAS_WRAPPER_BASE (internal variable)------
#-----------------------------------------------------
if [ "$JONAS_ROOT" = "$JONAS_BASE" ]; then
  JONAS_WRAPPER_BASE=$JONAS_WRAPPER_ROOT
else
  JONAS_WRAPPER_BASE=${JONAS_BASE}/wrapper
fi

if [ ! -d "$JONAS_WRAPPER_BASE" ]; then
  echo "Cannot find a valid JONAS_WRAPPER_BASE directory. Please put wrapper base directory in $JONAS_BASE/."
  exit 1
fi

# For Cygwin, switch paths to Windows format before running java
if $cygwin; then
  [ -n "$JONAS_WRAPPER_ROOT" ] &&
    JONAS_WRAPPER_ROOT=`cygpath --path --windows "$JONAS_WRAPPER_ROOT"`
  [ -n "$JONAS_WRAPPER_BASE" ] &&
    JONAS_WRAPPER_BASE=`cygpath --path --windows "$JONAS_WRAPPER_BASE"`
fi

#---------------------------------------------------
# Replace some wrapper conf properties (if not set)
#---------------------------------------------------
#replace token of jonas_wrapper_base.conf file
sed -i "/jonas.root/s@/path/to/jonas-root@$JONAS_ROOT@" ${JONAS_WRAPPER_BASE}/conf/jonas_wrapper_base.conf
sed -i "/jonas.base/s@/path/to/jonas-base@$JONAS_BASE@" ${JONAS_WRAPPER_BASE}/conf/jonas_wrapper_base.conf
sed -i "/wrapper.java.command/s@/path/to/java@${JAVA_HOME}/bin/java@" ${JONAS_WRAPPER_BASE}/conf/jonas_wrapper_base.conf

#replace token of jonas_start.conf & jonas_stop.conf files
sed -i "/wrapper.base/s@wrapper.base=/path/to/wrapper-base@wrapper.base=${JONAS_WRAPPER_BASE}@" ${JONAS_WRAPPER_BASE}/conf/jonas_start.conf
sed -i "/wrapper.base/s@wrapper.base=/path/to/wrapper-base@wrapper.base=${JONAS_WRAPPER_BASE}@" ${JONAS_WRAPPER_BASE}/conf/jonas_stop.conf

#replace wrapper_home and wrapper_base of setenv file
sed -i '/wrapper_root/s@"$PRGDIR"/..@'"${JONAS_WRAPPER_ROOT}"'@' ${JONAS_WRAPPER_BASE}/bat/setenv.sh
sed -i '/wrapper_base/s@"$wrapper_root"@'"${JONAS_WRAPPER_BASE}"'@' ${JONAS_WRAPPER_BASE}/bat/setenv.sh

#TODO before trying to replace a property, check that it's not already set.


#------------------------------------------------------------------
# Copy conf and bin files which'll be temporarily modified
#------------------------------------------------------------------
work=${JONAS_WRAPPER_BASE}/work

if [ ! -d "$work" ]; then
  mkdir ${JONAS_WRAPPER_BASE}/work
fi
if [ ! -d "$work"/conf ]; then
  mkdir ${JONAS_WRAPPER_BASE}/work/conf
fi
if [ ! -d "$work"/bin ]; then
  mkdir ${JONAS_WRAPPER_BASE}/work/bin
fi
cp ${JONAS_WRAPPER_BASE}/conf/jonas_start.conf ${JONAS_WRAPPER_BASE}/work/conf/
cp ${JONAS_WRAPPER_BASE}/conf/jonas_stop.conf ${JONAS_WRAPPER_BASE}/work/conf/
cp ${JONAS_WRAPPER_BASE}/conf/jonas_wrapper_base.conf ${JONAS_WRAPPER_BASE}/work/conf/
cp ${JONAS_WRAPPER_BASE}/bat/setenv.sh ${JONAS_WRAPPER_BASE}/work/bin/

cp ${JONAS_WRAPPER_BASE}/bat/installDaemon.sh ${JONAS_WRAPPER_BASE}/work/bin/
cp ${JONAS_WRAPPER_BASE}/bat/uninstallDaemon.sh ${JONAS_WRAPPER_BASE}/work/bin/
cp ${JONAS_WRAPPER_BASE}/bat/startDaemon.sh ${JONAS_WRAPPER_BASE}/work/bin/
cp ${JONAS_WRAPPER_BASE}/bat/stopDaemon.sh ${JONAS_WRAPPER_BASE}/work/bin/
cp ${JONAS_WRAPPER_BASE}/bat/queryDaemon.sh ${JONAS_WRAPPER_BASE}/work/bin/
cp ${JONAS_WRAPPER_BASE}/bat/runConsole.sh ${JONAS_WRAPPER_BASE}/work/bin/
cp ${JONAS_WRAPPER_BASE}/bat/wrapper.sh ${JONAS_WRAPPER_BASE}/work/bin/

#update links
sed -i '/conf_file/s@conf/@work/conf/@' ${JONAS_WRAPPER_BASE}/work/bin/setenv.sh
sed -i '/include/s@conf/@work/conf/@' ${JONAS_WRAPPER_BASE}/work/conf/jonas_start.conf
sed -i '/wrapper.stop.conf/s@conf/@work/conf/@' ${JONAS_WRAPPER_BASE}/work/conf/jonas_start.conf
sed -i '/include/s@conf/@work/conf/@' ${JONAS_WRAPPER_BASE}/work/conf/jonas_stop.conf


#-------------------------------------------------------------------
# Count the number of occurence of "wrapper.java.addition.x" token
#-------------------------------------------------------------------
COUNT=$(grep -c wrapper.java.additional. ${JONAS_WRAPPER_BASE}/work/conf/jonas_wrapper_base.conf)

# ---------------------------------------------
# Get arguments
# ---------------------------------------------
MODE=
JONAS_NAME=
USERNAME=
PASSWORD=
IS_STANDBY="false"

while [ "$#" -gt 0 ]
  do case "$1" in
    install)
      MODE=install
      ;;
    start)
      MODE=start
      ;;
    stop)
      MODE=stop
      ;;
    uninstall)
      MODE=uninstall
      ;;
    query)
      MODE=query
      ;;
    console)
      MODE=console
      ;;
    -standby)
      IS_STANDBY="true"
      ;;
    -running)
     # Implicit argument for the 'start' argument
      ;;
    -halt)
     # Implicit argument for the 'stop' argument
      ;;
    -cp)
      shift
      # insert a new classpath entry
      CLASSPATH_IND=$(grep -c wrapper.java.classpath. ${JONAS_WRAPPER_BASE}/work/conf/jonas_wrapper_base.conf)
      CLASSPATH_IND=$(($CLASSPATH_IND+1))
      echo "wrapper.java.additional.${CLASSPATH_IND} = $1" >> ${JONAS_WRAPPER_BASE}/work/conf/jonas_wrapper_base.conf
      ;;
    -debug)
      shift
      if [ "$1" != "-p" ]; then
        echo "-debug mode need a following \"-p \""
        exit 1
      fi
      shift
      JONAS_DEBUG_PORT=$1
      JONAS_DEBUG_SUSPEND=n
      if [ "$2" = "-s" ]; then
        shift
        shift
        JONAS_DEBUG_SUSPEND=$1
      fi
      echo "JOnAS Debug Info :"
      echo " listening on port : $JONAS_DEBUG_PORT"
      echo " suspend mode : $JONAS_DEBUG_SUSPEND"
      COUNT=$(($COUNT+1))
      echo "wrapper.java.additional.${COUNT} = -Xdebug" >> ${JONAS_WRAPPER_BASE}/work/conf/jonas_wrapper_base.conf
      COUNT=$(($COUNT+1))
      echo "wrapper.java.additional.${COUNT} = -Xnoagent" >> ${JONAS_WRAPPER_BASE}/work/conf/jonas_wrapper_base.conf
      COUNT=$(($COUNT+1))
      echo "wrapper.java.additional.${COUNT} = -Xrunjdwp:transport=dt_socket,server=y,address=$JONAS_DEBUG_PORT,suspend=$JONAS_DEBUG_SUSPEND" >> ${JONAS_WRAPPER_BASE}/work/conf/jonas_wrapper_base.conf
      ;;
    -nojit)
      COUNT=$(($COUNT+1))
      echo "wrapper.java.additional.${COUNT} = -Djava.compiler=NONE" >> ${JONAS_WRAPPER_BASE}/work/conf/jonas_wrapper_base.conf
      ;;
    -dev)
      COUNT=$(($COUNT+1))
      echo "wrapper.java.additional.${COUNT} = -Djonas.developer=true" >> ${JONAS_WRAPPER_BASE}/work/conf/jonas_wrapper_base.conf
      ;;
    -failStartOnError)
      COUNT=$(($COUNT+1))
      echo "wrapper.java.additional.${COUNT} = -Djonas.start.nofail=true" >> ${JONAS_WRAPPER_BASE}/work/conf/jonas_wrapper_base.conf
      ;;
    -clean)
      echo "Clean Felix cache before starting JOnAS."
      COUNT=$(($COUNT+1))
      echo "wrapper.java.additional.${COUNT} = -Djonas.cache.clean=true" >> ${JONAS_WRAPPER_BASE}/work/conf/jonas_wrapper_base.conf
      ;;
    -n)
      shift
      JONAS_NAME=$1
      if [ -n "$JONAS_NAME" ]; then
      	COUNT=$(($COUNT+1))
	    echo "wrapper.java.additional.${COUNT} = -Djonas.name=$JONAS_NAME" >> ${JONAS_WRAPPER_BASE}/work/conf/jonas_wrapper_base.conf
	  fi
      ;;
    -D*)
      COUNT=$(($COUNT+1))
      echo "wrapper.java.additional.${COUNT} = $1" >> ${JONAS_WRAPPER_BASE}/work/conf/jonas_wrapper_base.conf
      ;;
    -X*)
      COUNT=$(($COUNT+1))
      echo "wrapper.java.additional.${COUNT} = $1" >> ${JONAS_WRAPPER_BASE}/work/conf/jonas_wrapper_base.conf
      ;;
    -username)
      shift
      COUNT=$(($COUNT+1))
      echo "wrapper.java.additional.${COUNT} = -username $1" >> ${JONAS_WRAPPER_BASE}/work/conf/jonas_wrapper_base.conf
      ;;
    -password)
      shift
      COUNT=$(($COUNT+1))
      echo "wrapper.java.additional.${COUNT} = -password $1" >> ${JONAS_WRAPPER_BASE}/work/conf/jonas_wrapper_base.conf
      ;;
    *)
      # All other args are passed "as is" to the java program
      COUNT=$(($COUNT+1))
      echo "wrapper.java.additional.${COUNT} = $1" >> ${JONAS_WRAPPER_BASE}/work/conf/jonas_wrapper_base.conf
      ;;
  esac
  shift
done

# ---------------------------------------------
# Check args
# ---------------------------------------------
case "$MODE" in
  install)
    ;;
  start)
    #insert new standby argument option
    if [ "$IS_STANDBY" = "true" ]; then
	  echo "wrapper.app.parameter.2 = -standby" >> ${JONAS_WRAPPER_BASE}/work/conf/jonas_start.conf
    fi
    ;;
  stop)
    #insert new standby argument
    if [ "$IS_STANDBY" = "true" ]; then
      echo "wrapper.app.parameter.2 = -standby" >> ${JONAS_WRAPPER_BASE}/work/conf/jonas_stop.conf
    fi
    ;;
  uninstall)
    ;;
  query)
    ;;
  console)
    ;;
  *)
    # Usage
    # Will be displayed if the user type an unrecognized command
    echo "Unrecognized command : '$*'"
    echo "Usage: $0 install | start | stop | uninstall | query | console"
    echo "Options:"
    echo "  for 'start' or 'console': [-n] [-standby] [-debug] [-dev] [-clean] [-nojit]"
    echo "  -n <jonas-name>             Specify the JOnAS instance name."
    echo "  -standby                    Allow to reach the STANDBY state (Only applicable for 'start' and 'stop')."
    echo "  -debug -p <port> [-s <y/n>] Debug the JOnAS instance."
    echo "                              (allow to perform OSGi operations on the framework)."
    echo "                              (Only applicable for 'start')"
    echo "  -dev                        Use M2 registry for bundles (for JOnAS developers)."
    echo "  -failStartOnError           If the deployment of one deployable has failed when the server is starting, the startup of the server fail."
    echo "  -clean                      Clean Felix cache before starting JOnAS."
    echo "  -nojit                      Disable JIT."
    echo " for '*': [-cp] [-n] [-D*]"
    echo "  -cp <classpath>             Additional CLASSPATH entries (to be used with care)."
    echo "  -Dname=value                Add JVM properties."
    exit 1
    ;;
esac

# ---------------------------------------------
# Print out JONAS_BASE
# ---------------------------------------------
case "$MODE" in
 start)
 echo "JONAS_BASE is set to" $JONAS_BASE
 ;;
esac

#-----------------------------------------------------------------
# Install / Start / Stop / Uninstall / Query / Console  ----------
#-----------------------------------------------------------------
if [ "$MODE" = "install" ]; then
  exec ${JONAS_WRAPPER_BASE}/work/bin/installDaemon.sh
elif [ "$MODE" = "start" ]; then
  exec ${JONAS_WRAPPER_BASE}/work/bin/startDaemon.sh
elif [ "$MODE" = "stop" ]; then
  exec ${JONAS_WRAPPER_BASE}/work/bin/stopDaemon.sh
elif [ "$MODE" = "uninstall" ]; then
  exec ${JONAS_WRAPPER_BASE}/work/bin/uninstallDaemon.sh
elif [ "$MODE" = "query" ]; then
  exec ${JONAS_WRAPPER_BASE}/work/bin/queryDaemon.sh
elif [ "$MODE" = "console" ]; then
  exec ${JONAS_WRAPPER_BASE}/work/bin/runConsole.sh
fi